# Error index { {#--#} #error-index}

{# NOTE: Prevent Doxygen issue by writing the comment after the first header. #}
<!--
NOTE: This file has been generated automatically by “{{script}}”.
      Do not edit manually!
-->

This page lists the warnings and errors generated by xkbcommon.
There are currently {{ entries|length }} entries.

@todo The documentation of the log messages is a work in progress.

## Index {% raw %}{#error-code-list}{% endraw +%}

| Code      | Identifier                   | Description | Type |
| --------- | ---------------------------- | ----------- | ---- |
{% for entry in entries %}
| [{{entry.message_code}}] | `{{entry.id}}` | {{entry.description|prepend_todo}} | {{entry.type|capitalize}} |
{% endfor %}

## Details

{% for entry in entries %}
### {{entry.message_code}} – {{entry.message_name}} { {#--#}#{{entry.message_code}}}

<dl>
{% if entry.removed %}
  <dt>Added in</dt><dd>{{entry.added}}</dd>
  <dt>Removed in</dt><dd>{{entry.removed}}</dd>
{% else %}
  <dt>Since</dt><dd>{{entry.added}}</dd>
{% endif %}
  <dt>Type</dt><dd>{{entry.type|capitalize}}</dd>
  <dt>Summary</dt><dd>{{entry.description|prepend_todo}}</dd>
</dl>

{% if entry.details %}
{{entry.details}}

{% endif %}
{% if entry.examples %}
#### Examples

{% for example in entry.examples %}
<details>
  <summary>{{example.name}}</summary>

{{example.description}}
{% if example.before %}
**Fix:**
  <div class="example-container">
    <div class="example">
      <div class="example-inner">
        <div class="example-title">Before</div>
{{example.before-}}
      </div>
    </div>
    <div class="example">
      <div class="example-inner">
        <div class="example-title">After</div>
{{example.after-}}
      </div>
    </div>
  </div>
{% endif %}
</details>

{% endfor %}
{% endif %}
{% endfor %}
{% for entry in entries %}
[{{entry.message_code}}]: @ref {{entry.message_code}}
{% endfor %}
