/*
 * NOTE: This file has been generated automatically by “{{script}}”.
 *       Do not edit manually!
 */

/*
 * Copyright © 2024 Pierre Le Marre <dev@wismill.eu>
 * SPDX-License-Identifier: MIT
 */
#pragma once

#include "config.h"

#include "xkbcommon/xkbcommon.h"
#include "xkbcommon/xkbcommon-keysyms.h"

struct ambiguous_icase_ks_names_entry {
    xkb_keysym_t keysym;
    const int count;
    const char *names[{{ MAX_AMBIGUOUS_NAMES }}];
};

static const struct ambiguous_icase_ks_names_entry
ambiguous_icase_ks_names[] = {
    {% for names in ambiguous_case_insensitive_names.values() %}
    { XKB_KEY_{{ names[-1] }}, {{ names | length }}, { {# -#}
        {% for name in names -%}
        "{{ name }}"{% if not loop.last %}, {% endif -%}
        {% endfor %}
    {#- #} } }
    {%- if not loop.last %},{% endif +%}
    {% endfor %}
};
